# baleen.version
# Stores version information such that it can be read by setuptools.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Thu Feb 18 20:14:16 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: version.py [] benjamin@bengfort.com $

"""
Stores version information such that it can be read by setuptools.
"""

##########################################################################
## Imports
##########################################################################

__version_info__ = {
    'major': 0,
    'minor': 3,
    'micro': 0,
    'releaselevel': 'final',
    'serial': 0,
}


def get_version(short=False):
    """
    Computes a string representation of the version from __version_info__.
    """
    assert __version_info__['releaselevel'] in ('alpha', 'beta', 'final')
    vers = ["%(major)i.%(minor)i" % __version_info__, ]
    if __version_info__['micro']:
        vers.append(".%(micro)i" % __version_info__)
    if __version_info__['releaselevel'] != 'final' and not short:
        vers.append('%s%i' % (__version_info__['releaselevel'][0],
                              __version_info__['serial']))
    return ''.join(vers)



# baleen.console.app
# Definition of the Baleen Utility app and commands
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 10:54:51 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: app.py [] benjamin@bengfort.com $

"""
Definition of the Baleen Utility app and commands
http://bbengfort.github.io/tutorials/2016/01/23/console-utility-commis.html
"""

##########################################################################
## Imports
##########################################################################

from commis import color
from commis import ConsoleProgram
from commis.exceptions import ConsoleError

from baleen.console.commands import *
from baleen.version import get_version

##########################################################################
## Utility Definition
##########################################################################

DESCRIPTION = "Management and administration commands for Baleen"
EPILOG      = "If there are any bugs or concerns, submit an issue on Github"
COMMANDS    = [
    IngestCommand,
    ExportCommand,
    LoadOPMLCommand,
    SummaryCommand,
    RunCommand,
]


##########################################################################
## The Baleen CLI Utility
##########################################################################

class BaleenUtility(ConsoleProgram):

    description = color.format(DESCRIPTION, color.CYAN)
    epilog      = color.format(EPILOG, color.MAGENTA)
    version     = color.format("baleen v{}", color.CYAN, get_version())

    @classmethod
    def load(klass, commands=COMMANDS):
        utility = klass()
        for command in commands:
            utility.register(command)
        return utility



# baleen.console
# Implements the baleen console utility.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 10:52:36 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: __init__.py [] benjamin@bengfort.com $

"""
Implements the baleen console utility.
"""

##########################################################################
## Imports
##########################################################################

from .app import COMMANDS
from .app import BaleenUtility



# baleen.export
# Export an HTML corpus for analyses with NLTK
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Fri Oct 03 16:49:20 2014 -0400
#
# Copyright (C) 2014 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: export.py [] benjamin@bengfort.com $

"""
Export an HTML corpus for analyses with NLTK
"""

##########################################################################
## Imports
##########################################################################

import os
import codecs

from datetime import datetime
from baleen.models import Feed, Post

##########################################################################
## Exporter
##########################################################################

class MongoExporter(object):

    def __init__(self, categories=None):
        self.categories = categories

    @property
    def categories(self):
        if self._categories is None:
            self._categories = Feed.objects.distinct('category')
        return self._categories

    @categories.setter
    def categories(self, value):
        self._categories = value

    def feeds(self, categories=None):
        """
        Returns a list of feeds for the specified categories
        """
        if isinstance(categories, basestring):
            categories = [categories]
        elif categories is None:
            categories = self.categories

        for category in categories:
            for feed in Feed.objects(category=category):
                yield feed

    def posts(self, categories=None):
        """
        Returns a set of posts for the given category
        """
        for feed in self.feeds(categories):
            for post in Post.objects(feed=feed):
                yield post

    def readme(self):
        """
        Returns a string with the README to write to disk
        """

        total  = 0
        counts = []

        for cat in self.categories:
            count = sum(1 for p in self.posts(cat))
            counts.append("   %s: %i posts" % (cat, count))
            total += count
        counts = "\n".join(counts)

        return (
            "Baleen RSS Export\n"
            "=================\n\n"
            "These feeds were exported on %s\n\n"
            "There are %i posts in %i categories in this corpus as follows:\n%s\n"
        ) % (
            datetime.now().strftime("%b %d, %Y at %H:%M"),
            total,
            len(self.categories),
            counts
        )

    def export(self, root, categories=None):
        """
        In the root directory writes each file and a README
        """

        if not os.path.exists(root):
            os.mkdir(root)

        if not os.path.isdir(root):
            raise Exception("%s is not a directory!" % root)

        for category in self.categories:
            dirname = os.path.join(root, category.replace(" ", "_"))
            if not os.path.exists(dirname):
                os.mkdir(dirname)

            for idx, post in enumerate(self.posts(category)):
                name = os.path.join(dirname, "%03i.html" % idx)
                with codecs.open(name, 'wb', encoding='utf8') as f:
                    f.write(post.htmlize())

        readme = os.path.join(root, "README")
        with codecs.open(readme, 'wb', encoding='utf8') as f:
            f.write(self.readme())

if __name__ == '__main__':
    from baleen.models import connect
    connect()
    exporter = MongoExporter()
    exporter.export('fixtures/corpus')




# baleen.utils.logger
# Logging utility for Baleen
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Mon Sep 22 15:47:34 2014 -0400
#
# Copyright (C) 2014 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: logger.py [] benjamin@bengfort.com $

"""
Logging utility for Baleen
"""

##########################################################################
## Imports
##########################################################################

import logging
import getpass
import warnings
import logging.config

from baleen.config import settings
from baleen.utils.timez import COMMON_DATETIME

##########################################################################
## Logging configuration
##########################################################################

configuration = {
    'version': 1,
    'disable_existing_loggers': False,

    'formatters': {
        'simple': {
            'format': '%(name)s %(levelname)s [%(asctime)s] -- %(message)s',
            'datefmt': COMMON_DATETIME,
        }
    },

    'handlers': {
        'null': {
            'level': 'DEBUG',
            'class': 'logging.NullHandler',
        },

        'console': {
            'level': 'WARNING',
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
        },

        'logfile': {
            'level': 'INFO',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': settings.logfile,
            'maxBytes': '536870912', # 512 MB
            'formatter': 'simple',
        },

        'mongolog': {
            'level': 'INFO',
            'class': 'baleen.utils.mongolog.MongoHandler',
        }
    },

    'loggers': {
        'baleen': {
            'level': settings.loglevel,
            'handlers': ['logfile'],
            'propagagte': True,
        },
        'baleen.ingest': {
            'level': 'INFO',
            'handlers': ['logfile', 'mongolog'],
            'propagate': False,
        }
    },
}

logging.config.dictConfigClass(configuration).configure()
if not settings.debug: logging.captureWarnings(True)

##########################################################################
## Logger utility
##########################################################################

class WrappedLogger(object):
    """
    Wraps the Python logging module's logger object to ensure that all baleen
    logging happens with the correct configuration as well as any extra
    information that might be required by the log file (for example, the user
    on the machine, hostname, IP address lookup, etc).

    Subclasses must specify their logger as a class variable so all instances
    have access to the same logging object.
    """

    logger = None

    def __init__(self, **kwargs):
        self.raise_warnings = kwargs.pop('raise_warnings', settings.debug)
        self.logger = kwargs.pop('logger', self.logger)

        if not self.logger or not hasattr(self.logger, 'log'):
            raise TypeError(
                "Subclasses must specify a logger, not {}"
                .format(type(self.logger))
            )

        self.extras = kwargs

    def log(self, level, message, *args, **kwargs):
        """
        This is the primary method to override to ensure logging with extra
        options gets correctly specified.
        """
        extra = self.extras.copy()
        extra.update(kwargs.pop('extra', {}))

        kwargs['extra'] = extra
        self.logger.log(level, message, *args, **kwargs)

    def debug(self, message, *args, **kwargs):
        return self.log(logging.DEBUG, message, *args, **kwargs)

    def info(self, message, *args, **kwargs):
        return self.log(logging.INFO, message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        """
        Specialized warnings system. If a warning subclass is passed into
        the keyword arguments and raise_warnings is True - the warnning will
        be passed to the warnings module.
        """
        warncls = kwargs.pop('warning', None)
        if warncls and self.raise_warnings:
            warnings.warn(message, warncls)

        return self.log(logging.WARNING, message, *args, **kwargs)

    # Alias warn to warning
    warn = warning

    def error(self, message, *args, **kwargs):
        return self.log(logging.ERROR, message, *args, **kwargs)

    def critical(self, message, *args, **kwargs):
        return self.log(logging.CRITICAL, message, *args, **kwargs)


##########################################################################
## The Ingestion Logger Class
##########################################################################

class IngestLogger(WrappedLogger):
    """
    Performs logging for the baleen process with the log options above.
    """

    logger = logging.getLogger('baleen.ingest')

    def __init__(self, **kwargs):
        self._user = kwargs.pop('user', None)
        super(IngestLogger, self).__init__(**kwargs)

    @property
    def user(self):
        if not self._user:
            self._user = getpass.getuser()
        return self._user

    def log(self, level, message, *args, **kwargs):
        """
        Provide current user as extra context to the logger
        """
        extra = kwargs.pop('extra', {})
        extra.update({
            'user': self.user
        })

        kwargs['extra'] = extra
        super(IngestLogger, self).log(level, message, *args, **kwargs)


##########################################################################
## Logging Mixin
##########################################################################

class LoggingMixin(object):
    """
    Mix in to classes that need their own logging object!
    """

    @property
    def logger(self):
        """
        Instantiates and returns a IngestLogger instance
        """
        if not hasattr(self, '_logger') or not self._logger:
            self._logger = IngestLogger()
        return self._logger



# baleen.utils.timez
# Utility functions for Baleen
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Mon Sep 22 10:14:57 2014 -0400
#
# Copyright (C) 2014 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: utils.py [] benjamin@bengfort.com $

"""
Utility functions for Baleenc
"""

##########################################################################
## Imports
##########################################################################

import re
import time

from dateutil.tz import tzlocal, tzutc
from datetime import date, datetime, timedelta
from dateutil.relativedelta import relativedelta

##########################################################################
## Format constants
##########################################################################

HUMAN_DATETIME   = "%a %b %d %H:%M:%S %Y %z"
HUMAN_DATE       = "%b %d, %Y"
HUMAN_TIME       = "%I:%M:%S %p"
JSON_DATETIME    = "%Y-%m-%dT%H:%M:%S.%fZ" # Must be UTC
ISO8601_DATETIME = "%Y-%m-%dT%H:%M:%S%z"
ISO8601_DATE     = "%Y-%m-%d"
ISO8601_TIME     = "%H:%M:%S"
COMMON_DATETIME  = "%d/%b/%Y:%H:%M:%S %z"

##########################################################################
## Module helper function
##########################################################################

def localnow():
    return datetime.now(tzlocal())


def utcnow():
    now = datetime.utcnow()
    now = now.replace(tzinfo=tzutc())
    return now


zre = re.compile(r'([\-\+]\d{4})')
def strptimez(dtstr, dtfmt):
    """
    Helper function that performs the timezone calculation to correctly
    compute the '%z' format that is not added by default in Python 2.7.
    """
    if '%z' not in dtfmt:
        return datetime.strptime(dtstr, dtfmt)

    dtfmt  = dtfmt.replace('%z', '')
    offset = int(zre.search(dtstr).group(1))
    dtstr  = zre.sub('', dtstr)
    delta  = timedelta(hours = offset/100)
    utctsp = datetime.strptime(dtstr, dtfmt) - delta
    return utctsp.replace(tzinfo=tzutc())


def humanizedelta(*args, **kwargs):
    """
    Wrapper around dateutil.relativedelta (same construtor args) and returns
    a humanized string representing the detla in a meaningful way.
    """
    if 'milliseconds' in kwargs:
        sec  = kwargs.get('seconds', 0)
        msec = kwargs.pop('milliseconds')
        kwargs['seconds'] = sec + (float(msec) / 1000.0)

    delta = relativedelta(*args, **kwargs)
    attrs = ('years', 'months', 'days', 'hours', 'minutes', 'seconds')
    parts = [
        '%d %s' % (getattr(delta, attr), getattr(delta, attr) > 1 and attr or attr[:-1])
        for attr in attrs if getattr(delta, attr)
    ]

    return " ".join(parts)


##########################################################################
## Timer functions
##########################################################################


class Timer(object):
    """
    A context object timer. Usage:
        >>> with Timer() as timer:
        ...     do_something()
        >>> print timer.elapsed
    """

    def __init__(self, wall_clock=True):
        """
        If wall_clock is True then use time.time() to get the number of
        actually elapsed seconds. If wall_clock is False, use time.clock to
        get the process time instead.
        """
        self.wall_clock = wall_clock
        self.time = time.time if wall_clock else time.clock

        # Stubs for serializing an empty timer.
        self.started  = None
        self.finished = None
        self.elapsed  = 0.0

    def __enter__(self):
        self.started  = self.time()
        return self

    def __exit__(self, type, value, tb):
        self.finished = self.time()
        self.elapsed  = self.finished - self.started

    def __str__(self):
        return humanizedelta(seconds=self.elapsed)



# tests.utils_tests.test_logger
# Simple tests for the logger module.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Thu Mar 03 11:52:06 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: test_logger.py [] benjamin@bengfort.com $

"""
Simple tests for the logger module.
"""

##########################################################################
## Imports
##########################################################################

import getpass
import unittest

try:
    from unittest import mock
except ImportError:
    import mock

from baleen.utils.logger import *


##########################################################################
## Module Helpers
##########################################################################

PREFIX = "TEST LOG"
IGNORE = "IGNORE: This should not be in a log file or database!"

def tmsgf(message, prefix=PREFIX, ignore=IGNORE):
    return "{}: {} ({})".format(prefix, message, ignore)

##########################################################################
## Logger Test
##########################################################################

class IngestLoggerTests(unittest.TestCase):
    """
    Simply exercises the methods of the logger.
    """

    @mock.patch('baleen.utils.logger.IngestLogger.logger')
    def test_log_extra(self, mock_logger):
        """
        Assert that extra (user) is passed to logger
        """

        logger  = IngestLogger()

        assert logger.logger is mock_logger

        message = tmsgf("Do not double space after a period!")
        logger.log(logging.DEBUG, message)

        mock_logger.log.assert_called_with(logging.DEBUG, message, extra={'user': getpass.getuser()})

    @mock.patch('baleen.utils.logger.IngestLogger.logger')
    def test_log_debug(self, mock_logger):
        """
        Test the debug logger
        """

        logger  = IngestLogger()

        assert logger.logger is mock_logger

        message = tmsgf("All CAPS is not shouting!")
        logger.debug(message)

        mock_logger.log.assert_called_with(logging.DEBUG, message, extra=mock.ANY)

    @mock.patch('baleen.utils.logger.IngestLogger.logger')
    def test_log_info(self, mock_logger):
        """
        Test the info logger
        """

        logger  = IngestLogger()

        assert logger.logger is mock_logger

        message = tmsgf("Birds and Bees Flock with Seas!")
        logger.info(message)

        mock_logger.log.assert_called_with(logging.INFO, message, extra=mock.ANY)

    @mock.patch('baleen.utils.logger.IngestLogger.logger')
    def test_log_warn(self, mock_logger):
        """
        Test the warn logger
        """

        logger  = IngestLogger()

        assert logger.logger is mock_logger

        message = tmsgf("You shouldn't touch that hot stove!")
        logger.warn(message)

        mock_logger.log.assert_called_with(logging.WARNING, message, extra=mock.ANY)

    @mock.patch('baleen.utils.logger.IngestLogger.logger')
    def test_log_error(self, mock_logger):
        """
        Test the error logger
        """

        logger  = IngestLogger()

        assert logger.logger is mock_logger

        message = tmsgf("Someone let the rooster into the hen house!")
        logger.error(message)

        mock_logger.log.assert_called_with(logging.ERROR, message, extra=mock.ANY)

    @mock.patch('baleen.utils.logger.IngestLogger.logger')
    def test_log_critical(self, mock_logger):
        """
        Test the critical logger
        """

        logger  = IngestLogger()

        assert logger.logger is mock_logger

        message = tmsgf("Someone let the fox into the hen house!")
        logger.critical(message)

        mock_logger.log.assert_called_with(logging.CRITICAL, message, extra=mock.ANY)



# tests.utils_tests
# Tests for the Baleen utilities package.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sun Feb 21 15:31:55 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: __init__.py [] benjamin@bengfort.com $

"""
Tests for the Baleen utilities package.
"""

##########################################################################
## Imports
##########################################################################



# baleen.wrangle
# Wrangles the post objects from a synchronized feed.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 21:52:49 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: wrangle.py [] benjamin@bengfort.com $

"""
Wrangles the post objects from a synchronized feed.

Feed objects don't require a lot of wrangling, and are handled primarily by
the FeedSync object. However Posts do require some hoop jumping, which this
module provides.
"""

##########################################################################
## Imports
##########################################################################

import requests

from copy import deepcopy
from dateutil import parser as dtparser

from baleen.models import Post
from baleen.utils.decorators import reraise
from baleen.exceptions import WranglingError, FetchError

##########################################################################
## Module Constants
##########################################################################

FEEDPARSER_REMOVABLE_FIELDS = (
    'id', 'published_parsed', 'expired_parsed',
    'updated', 'updated_parsed',  'created', 'created_parsed',
)

##########################################################################
## Post Wrangling Object
##########################################################################

class PostWrangler(object):
    """
    As FeedSync wraps Feed to do work, so to does PostWrangler wrap an entry
    to create a Post object, to ensure that data is of a high quality, and to
    do extra things like fetch the full webpage from the URL provided.

    This object directly converts its input (a dict) to a models.Post object.
    """

    @classmethod
    def factory(klass, entries, feed=None):
        """
        Yields a post wrangler for each entry in the entries.
        """
        for entry in entries:
            yield klass(deepcopy(entry), feed=feed)

    def __init__(self, entry, feed=None):
        """
        Entry is expected to be the dictionary object from a FeedSync
        After wrangling, it will become a models.Post object.
        """
        self.feed = feed
        self.post = entry

    def is_wrangled(self):
        """
        Checks the class of the post to see if wrangling has occurred.
        """
        return isinstance(self.post, Post)

    @reraise(klass=WranglingError)
    def wrangle(self, save=True):
        """
        Converts the raw entry to standard data. If save, saves to database.

        Metholodolgy of wrangling is as follows:

            - all fields are kept in the entry except `published` and
              `published_parsed` since these many not contain TZ data -
              instead these two fields are replaced by `pubdate`. If there
              is no publication date, `pubdate` is set to None.

            - the tags field, if it exists, is converted to a list of
              strings. Although this may cause some data loss; it will
              make tagging of all posts simpler for the application.

            - link will be renamed url

            - content will be populated with summary, if content does not
              exist in the feed. Supposedly feedparser was already doing
              this, but it appears to not be regular.

            - title, url, content, and tags will all be encoded UTF-8.

            - removes the id field so a Mongo generated ObjectID is stored.

        See the models.Post for more information on the data structure.

        NOTE: This method is destructive, the raw entry will be converted.
        """
        ## Don't rewrangle an already wrangled post
        if self.is_wrangled():
            return self.post

        ## Saves typing self.post everywhere
        post = self.post.copy()

        ## Remove unwanted fields
        for field in FEEDPARSER_REMOVABLE_FIELDS:
            if field in post: del post[field]

        ## Handle the pubdate and published strings
        post['pubdate'] = dtparser.parse(post.pop('published')) if 'published' in post else None

        ## Handle the tags in the entry
        post['tags'] = [tag['term'] for tag in self.post.tags] if 'tags' in post else []

        ## Rename the link field to url
        post['url'] = self.post.link or post.get('href', None) or self.post.id
        if 'link' in post: del post['link']

        ## Handle the content
        if 'content' not in post:
            post['content'] = post.get('summary')
        else:
            selected = None
            for idx, item in enumerate(post['content']):
                if idx == 0:
                    # Take the first item
                    selected = item
                elif item['type'] == 'text/html':
                    # Unless we find another item that is html
                    selected = item

            # Update the post with the content info
            post['language'] = selected.get('language')
            post['mimetype'] = selected.get('type')
            post['content']  = selected.get('value')

        ## Create the post object
        ## Start using self.post here!
        self.post = Post(feed=self.feed, **post)
        if save:
            self.post.save()

        return self.post

    @reraise(klass=FetchError)
    def fetch(self, save=True):
        """
        Fetches the entire webpage for the post. If save, adds the page to
        the content of the post and saves it back to the database.

        Raises an exception if not wrangled yet.
        Raises exceptions if there is a problem with the fetch.
        """
        if not self.is_wrangled():
            raise ValueError("Entry not yet wrangled, cannot fetch.")

        response = requests.get(self.post.url)
        response.raise_for_status()

        if response.text:
            self.post.content = response.text

        if save:
            self.post.save()

        return self.post



# tests.test_wrangle
# Test the post wrangling module and functionality.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 22:38:08 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: .test_wrangle.py [] benjamin@bengfort.com $

"""
Test the post wrangling module and functionality.
"""

##########################################################################
## Imports
##########################################################################

import os
import pickle
import unittest

from .test_models import MongoTestMixin

try:
    from unittest import mock
except ImportError:
    import mock

from baleen.wrangle import *
from baleen.exceptions import *
from baleen.models import Feed, Post

##########################################################################
## Fixtures
##########################################################################

FIXTURES  = os.path.join(os.path.dirname(__file__), "fixtures")
RESULT    = os.path.join(FIXTURES, "feedparser_result.pickle")
FEED      = Feed(
    title = u'The Rumpus.net',
    link  = u'http://therumpus.net/feed/',
    urls  = {u'htmlurl': u'http://therumpus.net'}, category = u'books',
)


def mocked_requests_get(*args, **kwargs):
    class MockResponse:
        def __init__(self, text, status_code):
            self.text = text
            self.status_code = status_code

        def raise_for_status(self):
            if self.status_code != 200:
                raise Exception("HTTP {}".format(self.status_code))

    text = "Luke, I am your father!"

    if args[0] == 'http://example.com/vader/':
        return MockResponse(text, 200)

    return MockResponse("??", 404)

##########################################################################
## Test Wrangling Posts
##########################################################################

class PostWranglerTests(MongoTestMixin, unittest.TestCase):

    def setUp(self):
        super(PostWranglerTests, self).setUp()
        self.feed = FEED
        self.feed.save()

        with open(RESULT, 'rb') as f:
            self.entries = pickle.load(f).entries

    def test_wrangle_factory(self):
        """
        Test multiple types in the feed sync factory
        """

        for wrangle in PostWrangler.factory(self.entries, feed=self.feed):
            self.assertIsInstance(wrangle, PostWrangler)

    def test_wrangle_integration(self):
        """
        Test wrangling of all entries in the result.
        """
        self.assertEqual(Post.objects.count(), 0)
        for wrangle in PostWrangler.factory(self.entries, feed=self.feed):
            wrangle.wrangle()
            wrangle.wrangle() # Make sure that double wrangle does nothing.

        self.assertEqual(Post.objects.count(), 10)

        # Ensure there are no duplicates
        for wrangle in PostWrangler.factory(self.entries, feed=self.feed):
            with self.assertRaises(WranglingError) as cm:
                wrangle.wrangle()
            self.assertEqual(Post.objects.count(), 10)

    def test_is_wrangled(self):
        """
        Test the wrangling detection
        """
        wrangle = PostWrangler(self.entries[0])
        self.assertFalse(wrangle.is_wrangled())
        wrangle.wrangle()
        self.assertTrue(wrangle.is_wrangled())

    def test_save_not_save(self):
        """
        Test the wrangle interaction with the database
        """
        self.assertEqual(Post.objects.count(), 0)
        wrangle = PostWrangler(self.entries[0])

        # Don't save the wrangle
        wrangle.wrangle(False)
        self.assertEqual(Post.objects.count(), 0)

        # We've already wrangled so nothing should happen!
        wrangle.wrangle()
        self.assertEqual(Post.objects.count(), 0)

        # Try making something happen directly
        wrangle.wrangle().save()
        self.assertEqual(Post.objects.count(), 1)

        # Toss in something else entirely
        wrangle = PostWrangler(self.entries[1])
        wrangle.wrangle()
        self.assertEqual(Post.objects.count(), 2)

    def test_feed_or_not(self):
        """
        Test can be saved with or without a feed
        """
        withfeed = PostWrangler(self.entries[0], feed=self.feed)
        nofeed   = PostWrangler(self.entries[1])

        post = withfeed.wrangle()
        self.assertEqual(post.feed, self.feed)

        post = nofeed.wrangle()
        self.assertIsNone(post.feed)

    @mock.patch('baleen.wrangle.requests.get', side_effect=mocked_requests_get)
    def test_fetch_not_wrangled(self, mock_requests):
        """
        Assert that fetch requires wrangling
        """
        assert mock_requests is requests.get

        wrangle = PostWrangler(self.entries[0], feed=self.feed)
        with self.assertRaises(FetchError):
            wrangle.fetch()

    @mock.patch('baleen.wrangle.requests.get', side_effect=mocked_requests_get)
    def test_fetch_overwrites_content(self, mock_requests):
        """
        Test that the fetch overwrites content.
        """
        assert mock_requests is requests.get

        wrangle = PostWrangler(self.entries[0], feed=self.feed)
        wrangle.wrangle()
        self.assertEqual(Post.objects.count(), 1)

        wrangle.post.url = 'http://example.com/vader/'
        post = wrangle.fetch()
        self.assertEqual(Post.objects.count(), 1)
        self.assertNotEqual(post.created, post.updated)

        self.assertEqual(post.content, "Luke, I am your father!")

    @mock.patch('baleen.wrangle.requests.get', side_effect=mocked_requests_get)
    def test_fetch_no_save(self, mock_requests):
        """
        Test that the fetch does not save on demand.
        """
        assert mock_requests is requests.get

        wrangle = PostWrangler(self.entries[0], feed=self.feed)
        wrangle.wrangle()
        self.assertEqual(Post.objects.count(), 1)

        wrangle.post.url = 'http://example.com/vader/'
        wrangle.fetch(save=False)
        self.assertEqual(Post.objects.count(), 1)

        post = Post.objects.first()
        self.assertDateTimeEqual(post.created, post.updated)
        self.assertNotEqual(post.content, "Luke, I am your father!")

    @mock.patch('baleen.wrangle.requests.get', side_effect=mocked_requests_get)
    def test_fetch_raises_404(self, mock_requests):
        """
        Test that fetch raises exception on HTTP error
        """
        assert mock_requests is requests.get

        wrangle = PostWrangler(self.entries[0], feed=self.feed)
        wrangle.wrangle()
        self.assertEqual(Post.objects.count(), 1)

        with self.assertRaises(FetchError):
            wrangle.post.url = 'http://example.com/obiwan/'
            wrangle.fetch()
        



# baleen.console.utils
# Argparse extensions and utilities.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 11:01:35 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: baleen.console.utils.py [] benjamin@bengfort.com $

"""
Argparse extensions and utilities.
"""

##########################################################################
## Imports
##########################################################################

import argparse


##########################################################################
## Console Parsers
##########################################################################

def csv(type=int):
    """
    Argparse type for comma seperated values. Also parses the type, e.g. int.
    """
    def parser(s):
        try:
            parse = lambda p: type(p.strip())
            return map(parse, s.split(","))
        except Exception as e:
            raise argparse.ArgumentTypeError(
                "Could not parse CSV value to type {}: {!r}".format(type.__name__, s)
            )

    return parser



# test.utils_tests.test_timez
# Testing for the timez time helpers library.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sun Feb 21 15:33:18 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: test_timez.py [] benjamin@bengfort.com $

"""
Testing for the timez time helpers library.
"""

##########################################################################
## Imports
##########################################################################

import time
import unittest

from datetime import datetime
from dateutil.tz import tzutc
from baleen.utils.timez import *

##########################################################################
## Helper Functions Test Cases
##########################################################################

class TimezHelpersTests(unittest.TestCase):

    def setUp(self):
        self.localnow = datetime.now(tzlocal()).replace(microsecond=0)
        self.utcnow   = self.localnow.astimezone(tzutc())

    def tearDown(self):
        self.localnow = self.utcnow = None

    def test_non_naive_datetimes(self):
        """
        Assert that localnow and utcnow return non-naive datetimes
        """
        self.assertIsNotNone(localnow().tzinfo)
        self.assertIsNotNone(utcnow().tzinfo)

    def test_humanizedelta(self):
        """
        Test the humanize delta function to convert seconds
        """
        cases = (
            (12512334, "144 days 19 hours 38 minutes 54 seconds"),
            (34321, "9 hours 32 minutes 1 second"),
            (3428, "57 minutes 8 seconds"),
            (1, "1 second"),
            (0.21, "0 second"),
        )

        for seconds, expected in cases:
            self.assertEqual(humanizedelta(seconds=seconds), expected)

    def test_humanizedelta_milliseconds(self):
        """
        Test the humanize delta function to conver milliseconds
        """

        # Case with seconds already there
        self.assertEqual(humanizedelta(seconds=10, milliseconds=2000), '12 seconds')

        # Case without seconds present
        self.assertEqual(humanizedelta(milliseconds=456875), '7 minutes 36 seconds')

    def test_strptimez(self):
        """
        Test the parsing of timezone aware date strings
        """
        dtfmt = "%Y-%m-%dT%H:%M:%S%z"

        cases = (
            ('2012-12-27T12:53:12-0500', datetime(2012, 12, 27, 17, 53, 12, tzinfo=tzutc())),
            ('2012-12-27T12:53:12+0800', datetime(2012, 12, 27, 4, 53, 12, tzinfo=tzutc())),
        )

        for dtstr, dt in cases:
            self.assertEqual(dt, strptimez(dtstr, dtfmt))

        # Non-timezone case
        self.assertEqual(
            strptimez('2012-12-27T12:53:12', "%Y-%m-%dT%H:%M:%S"),
            datetime(2012, 12, 27, 12, 53, 12)
        )

    def test_strptimez_no_z(self):
        """
        Assert that strptimez works with no '%z'
        This should return a timezone naive datetime
        """
        dtfmt = "%a %b %d %H:%M:%S %Y"
        dtstr = self.localnow.strftime(dtfmt)
        self.assertEqual(strptimez(dtstr, dtfmt), self.localnow.replace(tzinfo=None))


    def test_strptimez_no_space(self):
        """
        Non-space delimited '%z' works
        """
        dtfmt = "%Y-%m-%dT%H:%M:%S%z"
        dtstr = self.localnow.strftime(dtfmt)
        self.assertEqual(strptimez(dtstr, dtfmt), self.utcnow)

    def test_begin_z(self):
        """
        Test fmt that begins with '%z'
        """
        dtfmt = "%z %H:%M:%S for %Y-%m-%d"
        dtstr = self.localnow.strftime(dtfmt)
        self.assertEqual(strptimez(dtstr, dtfmt), self.utcnow)

    def test_middle_z(self):
        """
        Test fmt that contains '%z'
        """
        dtfmt = "time is: %H:%M:%S %z on %Y-%m-%d "
        dtstr = self.localnow.strftime(dtfmt)
        self.assertEqual(strptimez(dtstr, dtfmt), self.utcnow)

    def test_timer(self):
        """
        Test the Timer context manager
        """
        with Timer() as t:
            time.sleep(1)

        self.assertGreater(t.finished, t.started)
        self.assertEqual(t.elapsed, t.finished-t.started)
        self.assertEqual(str(t), '1 seconds')



# baleen.console.commands.run
# Runs the ingestor in the background every hour.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 11:14:25 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: run.py [] benjamin@bengfort.com $

"""
Runs the ingestor in the background every hour.
"""

##########################################################################
## Imports
##########################################################################

import time
import baleen
import schedule
import baleen.models as db

from commis import Command
from functools import partial
from baleen.ingest import MongoIngestor
from baleen.utils.logger import IngestLogger

##########################################################################
## Command
##########################################################################

class RunCommand(Command):

    name = 'run'
    help = 'Runs the ingest command every hour'
    args = {}

    def ingest(self, args):
        db.connect()
        ingestor = MongoIngestor()
        ingestor.ingest()

    def handle(self, args):
        logger = IngestLogger()
        logger.info(
            "Starting Baleen v{} ingestion service every hour.".format(baleen.get_version())
        )

        schedule.every().hour.do(partial(self.ingest, args))

        while True:
            try:
                schedule.run_pending()
                time.sleep(1)
            except (KeyboardInterrupt, SystemExit):
                logger.info("Graceful shutdown of Baleen ingestion service.")
                return ""
            except Exception as e:
                logger.critical(str(e))
                return str(e)



# baleen.opml
# Reads opml files and gives back outline data
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sat Sep 20 23:12:07 2014 -0400
#
# Copyright (C) 2014 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: opml.py [] benjamin@bengfort.com $

"""
Reads opml files and gives back outline data
"""

##########################################################################
## Imports
##########################################################################

import baleen.models as db
from bs4 import BeautifulSoup
from collections import Counter
from mongoengine.errors import *

##########################################################################
## Load Database function
##########################################################################

def load_opml(path):
    """
    Loads an OPML file into the Mongo database; returns the count of the
    number of documents added to the database.
    """

    opml = OPML(path)
    rows = 0
    for feed in opml:
        feed.pop('type')                    # Unneeded for database
        feed.pop('text')                    # Unneeded for database
        feed['link'] = feed.pop('xmlUrl')   # Rename the XML URL
        feed['urls'] = {
            'xmlUrl':  feed['link'],        # Add xmlUrl to urls
            'htmlUrl': feed.pop('htmlUrl'), # Add htmlUrl to urls
        }
        feed = db.Feed(**feed)              # Construct without an ObjectId

        try:
            feed.save()
            rows += 1
        except NotUniqueError:
            continue
    return rows

##########################################################################
## OPMLReader
##########################################################################

class OPML(object):

    def __init__(self, path):
        """
        Reader for OPML XML files.
        """
        self.path = path

    def categories(self):
        """
        Reads the file to capture all the categories
        """
        with open(self.path, 'r') as data:
            soup = BeautifulSoup(data, 'xml')
            for topic in soup.select('body > outline'):
                yield topic['title']

    def counts(self):
        """
        Returns the counts of feeds in each category
        """
        counts = Counter()
        for item in self:
            counts[item['category']] += 1
        return counts

    def __iter__(self):
        """
        Yields a dictionary representing the attributes of the RSS feed
        from the OPML file; also captures category data.
        """
        with open(self.path, 'r') as data:
            soup = BeautifulSoup(data, 'xml')
            for topic in soup.select('body > outline'):
                for feed in topic.find_all('outline'):
                    data = feed.attrs.copy()
                    data['category'] = topic['title']
                    yield data

    def __len__(self):
        return sum(1 for item in self)

    def __str__(self):
        counts = self.counts()
        return "OPML with {} categories and {} feeds".format(
            len(counts), sum(counts.values())
        )

    def __repr__(self):
        return "<{} at {}>".format(self.__class__.__name__, self.path)



# test.test_utils.test_decorators
# Testing the decorators utility package.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 19:06:34 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: test_decorators.py [] benjamin@bengfort.com $

"""
Testing the decorators utility package.
"""

##########################################################################
## Imports
##########################################################################

import time
import unittest

from baleen.utils.decorators import *
from baleen.utils.timez import Timer
from baleen.exceptions import *

try:
    from unittest import mock
except ImportError:
    import mock


##########################################################################
## Decorators Tests
##########################################################################

class DecoratorsTests(unittest.TestCase):
    """
    Basic decorators utility tests.
    """

    def test_memoized(self):
        """
        Test the memoized property
        """

        class Thing(object):

            @memoized
            def attr(self):
                return 42

        thing = Thing()
        self.assertFalse(hasattr(thing, '_attr'))
        self.assertEqual(thing.attr, 42)
        self.assertTrue(hasattr(thing, '_attr'))

    def test_timeit(self):
        """
        Test the timeit decorator
        """

        @timeit
        def myfunc():
            return 42

        output = myfunc()
        self.assertEqual(len(output), 2)
        result, timer = output
        self.assertEqual(result, 42)
        self.assertTrue(isinstance(timer, Timer))

    def test_reraise(self):
        """
        Test the reraise decorator
        """

        # Test 1: Regular old reraise

        @reraise()
        def alpha():
            raise Exception("Should be a BaleenError")

        with self.assertRaises(BaleenError) as cm:
            alpha()

        e = cm.exception
        self.assertEqual(str(e), "Should be a BaleenError")
        self.assertTrue(hasattr(e, "original"))
        self.assertIsInstance(e.original, Exception)
        self.assertEqual(str(e.original), "Should be a BaleenError")

    def test_reraise_message(self):
        """
        Test the reraise decorator with a message
        """

        # Test 2: Reraise with a new message

        @reraise(message="I'm handling it!")
        def bravo():
            raise NotImplementedError("I'm not handling it!")

        with self.assertRaises(BaleenError) as cm:
            bravo()

        e = cm.exception
        self.assertEqual(str(e), "I'm handling it!")
        self.assertTrue(hasattr(e, "original"))
        self.assertIsInstance(e.original, NotImplementedError)
        self.assertEqual(str(e.original), "I'm not handling it!")

    def test_reraise_arguments(self):
        """
        Test the reraise decorator with all possible arguments
        """

        # Test 3: All possible arguments to reraise

        @reraise(klass=FeedTypeError, message="bad feed type", trap=TypeError)
        def charlie():
            raise TypeError("requires an integer")

        with self.assertRaises(FeedTypeError) as cm:
            charlie()

        e = cm.exception
        self.assertEqual(str(e), "bad feed type")
        self.assertTrue(hasattr(e, "original"))
        self.assertIsInstance(e.original, TypeError)
        self.assertEqual(str(e.original), "requires an integer")

    def test_reraise_trap(self):
        """
        Test the reraise decorator by missing the trap
        """

        # Test 4: Missing the trap

        @reraise(klass=FeedTypeError, message="bad feed type", trap=TypeError)
        def delta():
            raise ValueError("this should be the exception raised")

        with self.assertRaises(ValueError) as cm:
            delta()

        e = cm.exception
        self.assertEqual(str(e), "this should be the exception raised")
        self.assertFalse(hasattr(e, "original"))



# baleen.exceptions
# Exceptions hierarchy for the Baleen module.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 13:59:03 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: exceptions.py [] benjamin@bengfort.com $

"""
Exceptions hierarchy for the Baleen module.
"""

##########################################################################
## Exceptions Hierarchy
##########################################################################

class BaleenError(Exception):
    """
    The root of all errors in Baleen (hopefully)
    """
    pass


class FeedTypeError(BaleenError):
    """
    Could not detect the feed type for synchronization
    """
    pass


class IngestionError(BaleenError):
    """
    Something went wrong specifically with the ingestion process
    """
    pass


class SynchronizationError(IngestionError):
    """
    Something went wrong with feed synchronization
    """
    pass


class WranglingError(IngestionError):
    """
    Something went wrong wrangling a post
    """
    pass


class FetchError(WranglingError):
    """
    Could not fetch the webpage for the post
    """
    pass



# tests.utils_tests.test_mongolog
# Simple tests for logging to MongoDB
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Thu Mar 03 11:53:46 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: test_mongolog.py [] benjamin@bengfort.com $

"""
Simple tests for logging to MongoDB
"""

##########################################################################
## Imports
##########################################################################

import logging
import unittest

from mongomock import MongoClient as MockMongoClient

try:
    from unittest import mock
except ImportError:
    import mock

from baleen.utils import mongolog as ml
from .test_logger import tmsgf

##########################################################################
## Mongo Log Handler Tests
##########################################################################

class MongoLogHandlerTests(unittest.TestCase):
    """
    Simply exercises the methods of the logger.
    """

    @mock.patch('baleen.utils.mongolog.MongoClient', MockMongoClient)
    def test_logging_to_mongo(self):
        """
        Test the mongo log handler and logging to mongo
        """
        assert ml.MongoClient is MockMongoClient

        handler = ml.MongoHandler(level=logging.DEBUG)
        self.assertIsInstance(handler.connection, MockMongoClient)

        # Ensure there is nothing in the database.
        self.assertEqual(handler.collection.count(), 0)

        # Create the logging instance.
        logger = logging.getLogger('test.mongo.logger.demo')
        logger.setLevel(logging.INFO)
        logger.addHandler(handler)

        # Log a message
        logger.info(tmsgf("This is a test of the mongo logger"))

        # Ensure there is now a log message
        self.assertEqual(handler.collection.count(), 1)



# baleen.console.commands
# Comamnds for the Baleen CLI utility.
#
# Author:   Benjamin Bengfort <bengfort@cs.umd.edu>
# Created:  Wed Mar 02 10:54:07 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: __init__.py [] benjamin@bengfort.com $

"""
Comamnds for the Baleen CLI utility.
"""

##########################################################################
## Imports
##########################################################################

from .ingest import IngestCommand
from .export import ExportCommand
from .load import LoadOPMLCommand
from .summary import SummaryCommand
from .run import RunCommand



# baleen.ingest
# The ingestion runner that implements ingestion for a collection of feeds.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 23:23:06 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: ingest.py [] benjamin@bengfort.com $

"""
The ingestion runner that implements ingestion for a collection of feeds.
"""

##########################################################################
## Imports
##########################################################################

import uuid

from baleen.opml import OPML
from baleen.exceptions import *
from baleen import models as db
from baleen.feed import FeedSync
from baleen.config import settings
from baleen.utils.timez import Timer
from baleen.wrangle import PostWrangler
from baleen.utils.logger import LoggingMixin
from baleen.utils.decorators import memoized

from datetime import datetime
from collections import Counter


##########################################################################
## Helper Functions
##########################################################################

def stype(obj):
    """
    Returns the string of the type. Used to count exception types.
    """
    if isinstance(obj, BaleenError):
        if hasattr(obj, "original"):
            return "{} ({})".format(
                type(obj).__name__, type(obj.original).__name__
            )
    return type(obj).__name__


##########################################################################
## Base Ingestion Class
##########################################################################

class Ingestor(LoggingMixin):
    """
    Base class for the ingestors.

    Ingestors manage the synchronization of feeds, wrangling of posts, and
    fetching of web pages to store to the Mongo database. Ingestors can
    either get feeds from a list of strings, an OPML file or a Mongo query.

    Ingestors also perform logging and exception handling.
    """

    def __init__(self, feeds=None, **options):
        self.timer   = None         # Processing timer
        self.jobid   = None         # Unique job id for every run
        self.options = options      # Any other options passed in
        self._feeds  = feeds        # Allows pass in feed collection
        self.errors  = Counter()    # Count the number of error types

    @property
    def name(self):
        return self.__class__.__name__

    @memoized
    def counts(self):
        """
        Keep track of counts and ensure zero keys exist.
        """
        counts = Counter()
        for key in ('feeds', 'posts', 'errors', 'feed_error'):
            counts[key] = 0
        return counts

    def feeds(self):
        """
        This is the primary entry point for subclasses, they must specificy
        how to get access to a collection of feeds to syncrhonize.
        """
        if self._feeds is not None:
            return self._feeds

        raise IngestionError(
            "No feeds specified for {} ingestion!".format(self.name)
        )

    def started(self):
        """
        Run when the ingestor is started and used for logging. Subclasses can
        use it as a hook to perform extra work right before kick off.
        """
        message = "{} job {} started".format(self.name, self.jobid)
        self.logger.info(message)

    def failed(self, exception):
        """
        Executed when a complete ingestion run has failed (very bad). Used
        to log the exception or clean up before Baleen crashes!
        """
        message = "{} job {} failed!".format(self.name, self.jobid)
        self.logger.error("Ingestion Error: {}".format(exception))
        self.logger.critical(message)

    def finished(self):
        """
        Run when the ingestor has finished and used for logging. Subclasses
        can use it as a hook to perform any completion work.
        """
        # Notify the results
        results = (
            "Processed {feeds} feeds ({timer}) "
            "{posts} posts with {errors} errors"
        ).format(
            timer=self.timer, **self.counts
        )
        self.logger.info(results)

        # Notify job finished
        message  = "{} job {} finished".format(self.name, self.jobid)
        self.logger.info(message)

    def process(self):
        """
        Runs the ingestion process by iterating over the feeds, synchronizing
        and then wrangling posts into the database as well as fetching pages.
        """
        for idx, fsync in enumerate(FeedSync.factory(self.feeds())):
            try:
                self.process_feed(fsync)
                self.counts['feeds'] += 1
            except SynchronizationError as e:
                self.counts['feed_error'] += 1
                self.errors[stype(e)] += 1
                self.logger.error(
                    u"Error on Feed {} ({}): {}".format(
                        idx+1, fsync.feed, unicode(e)
                    )
                )

    def process_feed(self, fsync):
        """
        Synchronizes a feed and catches exceptions
        """
        factory = PostWrangler.factory(fsync.entries(), fsync.feed)
        for idx, post in enumerate(factory):
            try:
                self.process_post(post)
                self.counts["posts"] += 1
            except WranglingError as e:
                self.counts["errors"] += 1
                self.errors[stype(e)] += 1
                self.logger.error(
                    u"Post Error for feed {} on entry {}: {}".format(
                        fsync.feed, idx, unicode(e)
                    )
                )

    def process_post(self, post):
        """
        Wrangles a post from a single feed and catches exceptions
        """
        post.wrangle()
        if settings.fetch_html:
            try:
                post.fetch()
            except FetchError as e:
                self.counts["fetch_error"] += 1
                self.errors[stype(e)] += 1
                self.logger.error(
                    u"Fetch Error for post \"{}\" ({}): {}".format(
                        post.post.title, post.post.url, unicode(e)
                    )
                )

    def ingest(self):
        """
        Subclasses do not typically override the ingest method. Instead they
        will override the process hooks for start, failed, and finish,  or the
        process method directly.
        """
        # Set a unique job id for every time run is called.
        # The job id is based on the hostname and a time sequence.
        self.jobid = uuid.uuid1()

        # Call the started hook for logging and notification.
        self.started()

        # Time how long it takes to perform the processing
        with Timer() as self.timer:
            try:
                self.process()
            except Exception as e:
                # If something goes wrong, call the failed hook, then raise.
                self.failed(e)
                raise

        # Call the finished hook for logging and notification.
        self.finished()


##########################################################################
## Mongo Ingestion Class
##########################################################################

class MongoIngestor(Ingestor):
    """
    Ingests feeds that are stored in the database.
    This type of ingestor also tracks information into the database.
    """

    def feeds(self):
        """
        Returns an iterator of all active feeds from the database
        """
        for feed in db.Feed.objects(active=True):
            yield feed

    def started(self):
        """
        Save a record about the job start to the database.
        """
        super(MongoIngestor, self).started()
        self.job = db.Job(jobid=self.jobid, name=self.name)
        self.job.save()

    def failed(self, exception):
        """
        Save information about the failure to the database.
        """
        super(MongoIngestor, self).failed(exception)
        self.job.failed = True
        self.job.reason = unicode(exception)
        self.job.finished = datetime.now()
        self.job.save()

    def finished(self):
        """
        Update the job record in the database.
        """
        super(MongoIngestor, self).finished()
        self.job.reason = u"OK"
        self.job.finished = datetime.now()
        self.job.counts = self.counts
        self.job.errors = self.errors
        self.job.totals = {
            "feeds": db.Feed.objects.count(),
            "posts": db.Post.objects.count(),
            "jobs": db.Job.objects.count(),
        }
        self.job.save()

##########################################################################
## OPML Ingestion Class
##########################################################################

class OPMLIngestor(Ingestor):
    """
    Ingests feeds from an OPML file.
    """

    def __init__(self, path, **options):
        self.opml = OPML(path)
        super(OPMLIngestor, self).__init__(**options)

    def feeds(self):
        """
        Returns an iterator of all active feeds from the database
        """
        for feed in self.opml:
            yield feed



# baleen.console.commands.ingest
# Handles the ingestion utility both for OPML and feeds.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 10:58:56 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: ingest.py [] benjamin@bengfort.com $

"""
Handles the ingestion utility both for OPML and feeds.
"""

##########################################################################
## Imports
##########################################################################

import baleen.models as db

from commis import Command
from commis.exceptions import ConsoleError
from baleen.ingest import Ingestor, MongoIngestor, OPMLIngestor

##########################################################################
## Command
##########################################################################

class IngestCommand(Command):

    name = 'ingest'
    help = 'Ingests the RSS feeds to MongoDB'
    args = {
        '--opml': {
            'type': str,
            'default': None,
            'help': 'Ingest directly from an OPML file',
        },
        'feeds': {
            'type': str,
            'nargs': "*",
            'default': None,
            'metavar': 'URL',
            'help': 'Specify a list of feeds as urls'
        }
    }

    def handle(self, args):

        ingestor = MongoIngestor()

        if args.opml:
            ingestor = OPMLIngestor(args.opml)
            raise ConsoleError("opml ingestion is an untested utility!")

        if args.feeds:
            ingestor = Ingestor(args.feeds)
            raise ConsoleError("feed ingestion is an untested utility!")

        db.connect()
        ingestor.ingest()
        return (
            "Processed {feeds} feeds ({timer}): "
            "{posts} posts with {errors} errors"
        ).format(
            timer=ingestor.timer, **ingestor.counts
        )



# baleen.utils.decorators
# Decorators and function utilities for Baleen.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 19:03:43 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: decorators.py [] benjamin@bengfort.com $

"""
Decorators and function utilities for Baleen.
"""

##########################################################################
## Imports
##########################################################################

import time

from functools import wraps
from baleen.utils.timez import Timer
from baleen.exceptions import BaleenError

##########################################################################
## Memoization
##########################################################################

def memoized(fget):
    """
    Return a property attribute for new-style classes that only calls its
    getter on the first access. The result is stored and on subsequent
    accesses is returned, preventing the need to call the getter any more.
    https://github.com/estebistec/python-memoized-property
    """
    attr_name = '_{0}'.format(fget.__name__)

    @wraps(fget)
    def fget_memoized(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fget(self))
        return getattr(self, attr_name)

    return property(fget_memoized)


##########################################################################
## Timer functions
##########################################################################

def timeit(func):
    """
    Returns the number of seconds that a function took along with the result
    """

    @wraps(func)
    def timer_wrapper(*args, **kwargs):
        """
        Inner function that uses the Timer context object
        """
        with Timer() as timer:
            result = func(*args, **kwargs)

        return result, timer

    return timer_wrapper


##########################################################################
## Exception Handling
##########################################################################

def reraise(klass=BaleenError, message=None, trap=Exception):
    """
    Catches exceptions (those specified by trap) and then reraises the
    exception type specified by class. Also embeds the original exception as
    a property of the new exception: `error.original`. Finally you can
    specify another message to raise, otherwise the error string is used.
    """

    def reraise_decorator(func):

        @wraps(func)
        def reraise_wrapper(*args, **kwargs):
            """
            Capture Wrapper
            """
            try:
                return func(*args, **kwargs)
            except trap as e:
                error = klass(message or e.message)
                error.original = e
                raise error

        return reraise_wrapper

    return reraise_decorator



# baleen.console.commands.summary
# A utility to print out information about the Baleen state.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 11:08:57 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: summary.py [] benjamin@bengfort.com $

"""
A utility to print out information about the Baleen state.
"""

##########################################################################
## Imports
##########################################################################

import baleen
import baleen.models as db

from commis import Command
from baleen.config import settings
from baleen.utils.timez import HUMAN_DATETIME

##########################################################################
## Command
##########################################################################

class SummaryCommand(Command):

    name = 'info'
    help = 'Print info about Baleen from the database'
    args = {
        ('-c', '--config'): {
            'action': 'store_true',
            'default': False,
            'help': 'Also print the configuration',
        }
    }

    def handle(self, args):
        # Setup output and connect to database.
        output = []
        db.connect()

        # Printout configuration details as necessary.
        if args.config:
            output.append(u"Configuration:")
            output.append(unicode(settings))
            output.append(u"")

        output.append(u"Baleen v{} Status:".format(baleen.get_version()))
        output.append(
            u"{} Feeds and {} Posts after {} Jobs".format(
                db.Feed.objects.count(),
                db.Post.objects.count(),
                db.Job.objects.count(),
            )
        )

        latest = db.Job.objects.order_by('-started').first()
        output.extend([
            u"",
            u"Latest Job: ",
            u"    Type: {} v{}".format(latest.name, latest.version),
            u"    Job ID: {}".format(latest.jobid),
            u"    Started: {}".format(latest.started.strftime(HUMAN_DATETIME))
        ])

        if latest.finished:
            if latest.failed:
                output.append(u"    Failed: {}".format(latest.reason))
            else:
                output.append(u"    Finished: {}".format(latest.finished.strftime(HUMAN_DATETIME)))
                output.append(u"    Counts:")
                output.append(u"      " + u"\n      ".join([u"{}: {}".format(*item) for item in latest.counts.items()]))
                output.append(u"    Errors:")
                output.append(u"      " + u"\n      ".join([u"{}: {}".format(*item) for item in latest.errors.items()]))
        else:
            output.append(u"    Currently Running")

        latest = db.Feed.objects.order_by('-updated').first()
        output.extend([
            u"",
            u"Latest Feed: ",
            u"    Title: \"{}\"".format(latest.title),
            u"    eTag: \"{}\"".format(latest.etag),
            u"    Modified: {}".format(latest.modified),
            u"    Updated: {}".format(latest.updated.strftime(HUMAN_DATETIME)),
            u"    Posts: {}".format(latest.count_posts()),
        ])

        latest = db.Post.objects.order_by('-id').first()
        output.extend([
            u"",
            u"Latest Post: ",
            u"    Title: \"{}\"".format(latest.title),
            u"    Feed: \"{}\"".format(latest.feed.title),
            u"    Fetched: {}".format(latest.created.strftime(HUMAN_DATETIME)),
        ])

        return u"\n".join(output).encode('utf-8', errors='replace')



# baleen.console.commands.load
# Loads an OPML file from disk into the database.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 11:05:57 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: load.py [] benjamin@bengfort.com $

"""
Loads an OPML file from disk into the database.
"""

##########################################################################
## Imports
##########################################################################

from commis import Command
from baleen import models as db
from baleen.opml import load_opml

##########################################################################
## Command
##########################################################################

class LoadOPMLCommand(Command):

    name = 'load'
    help = 'Loads an OPML file from disk into the database.'
    args = {
        'opml': {
            'nargs': "+",
            'type': str,
            'help': 'OPML file(s) to import to the database'
        }
    }

    def handle(self, args):
        # Connect to the database
        db.connect()

        # Load the OPML files into the database
        count = sum(load_opml(path) for path in args.opml)
        return "Ingested {} feeds from {} OPML files".format(count, len(args.opml))



# tests.test_opml
# Testing for the OPML reader and ingestion function.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Fri Feb 19 08:50:19 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: test_opml.py [] benjamin@bengfort.com $

"""
Testing for the OPML reader and ingestion function.
"""

##########################################################################
## Imports
##########################################################################

import os
import unittest

from .test_models import MongoTestMixin

try:
    from unittest import mock
except ImportError:
    import mock

from baleen.opml import OPML, load_opml
from baleen.models import Feed

##########################################################################
## Fixtures
##########################################################################

FIXTURES = os.path.join(os.path.dirname(__file__), "fixtures")
FEEDLY   = os.path.join(FIXTURES, "feedly.opml")

##########################################################################
## Test Load OPML command
##########################################################################

class LoadOPMLTests(MongoTestMixin, unittest.TestCase):

    def test_load_opml_integrated(self):
        """
        Test the integration of the ingest helper function
        """
        self.assertEqual(Feed.objects.count(), 0)
        self.assertEqual(load_opml(FEEDLY), 36)
        self.assertEqual(Feed.objects.count(), 36)

        for feed in Feed.objects():
            self.assertIn('xmlUrl', feed.urls)
            self.assertIn('htmlUrl', feed.urls)

    def test_load_opml_no_duplicates(self):
        """
        Assert multiple calls to the load_opml creates no duplicates
        """
        self.assertEqual(Feed.objects.count(), 0)
        self.assertEqual(load_opml(FEEDLY), 36)
        self.assertEqual(Feed.objects.count(), 36)

        for _ in xrange(10):
            self.assertEqual(load_opml(FEEDLY), 0)
            self.assertEqual(Feed.objects.count(), 36)

##########################################################################
## OPML Reader Test
##########################################################################

class OPMLTests(unittest.TestCase):

    def test_fixture(self):
        """
        Assert the required opml fixture is available
        """
        self.assertTrue(os.path.exists(FEEDLY))
        self.assertTrue(os.path.isfile(FEEDLY))

    def test_categories(self):
        """
        Test the OPML categories listing
        """
        opml = OPML(FEEDLY)
        expected = [
            u'news',
            u'do it yourself',
            u'business',
            u'gaming',
            u'data science',
            u'essays',
            u'politics',
            u'tech',
            u'cinema',
            u'books',
            u'sports',
            u'cooking',
            u'design'
        ]

        print list(opml.categories())

        self.assertEqual(list(opml.categories()), expected)

    def test_length(self):
        """
        Test the OPML len built in
        """
        opml = OPML(FEEDLY)
        self.assertEqual(len(opml), 36)

    def test_counts(self):
        """
        Test the OPML category counter and item iterator
        """
        opml = OPML(FEEDLY)
        expected = {
            'cooking': 4,
            'cinema': 3,
            'gaming': 3,
            'tech': 3,
            'essays': 2,
            'business': 3,
            'design': 2,
            'sports': 3,
            'books': 3,
            'data science': 4,
            'do it yourself': 2,
            'news': 2,
            'politics': 2,
        }
        counts = opml.counts()

        for key, val in expected.items():
            self.assertIn(key, counts)
            self.assertEqual(
                counts[key], val,
                "{} mismatch: {} vs {}".format(key, counts[key], val)
            )

    def test_item_iterator_detail(self):
        """
        Test the XML result returned from OPML iteration
        """

        opml  = OPML(FEEDLY)
        attrs = ['category', 'title', 'text', 'htmlUrl', 'xmlUrl', 'type']
        for item in opml:
            self.assertTrue(isinstance(item, dict))
            self.assertEqual(item.keys(), attrs)



# tests.test_ingest
# Test the ingestor mechanism in an integration fashion.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Thu Mar 03 13:01:12 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: test_ingest.py [] benjamin@bengfort.com $

"""
Test the ingestor mechanism in an integration fashion.
"""

##########################################################################
## Imports
##########################################################################

import unittest

from .test_models import MongoTestMixin

try:
    from unittest import mock
except ImportError:
    import mock

import baleen.models as db

from baleen.ingest import stype
from baleen.ingest import Ingestor
from baleen.ingest import MongoIngestor
from baleen.ingest import OPMLIngestor
from baleen.utils.decorators import reraise
from baleen.exceptions import *
from baleen.utils.logger import IngestLogger


##########################################################################
## Helper Functions
##########################################################################

ACTION_METHODS = ('started', 'finished', 'failed', 'process')

def get_ingest_mock(klass=Ingestor):
    """
    Mocks all functions of the ingestor that are called in ingest.
    This means there should be NO side effects when ingest is called.
    """
    # Verify and create ingestor class
    ingestor = klass()
    verify_ingest_mock(ingestor)

    # Remove action methods
    for method in ACTION_METHODS:
        setattr(ingestor, method, mock.MagicMock())

    return ingestor


def verify_ingest_mock(ingestor):
    """
    Ensures that no methods other than action methods are called
    """
    ingestor = mock.create_autospec(ingestor, instance=True)
    reset_mock_method(ingestor, 'ingest')
    ingestor.ingest()

    for method in ingestor._mock_methods:
        action = getattr(ingestor, method)
        if method not in ACTION_METHODS:
            if hasattr(action, 'assert_not_called'):
                action.assert_not_called()


def reset_mock_method(obj, method):
    """
    Resets a mock object's method to the orignal
    """
    klass  = obj.__class__
    action = getattr(klass, method)

    setattr(obj, method, action.__get__(obj, klass))
    return obj


##########################################################################
## Test Ingestor
##########################################################################

class IngestorTests(MongoTestMixin, unittest.TestCase):

    def test_stype_helper(self):
        """
        Test the stype helper function
        """
        self.assertEqual(stype(BaleenError("Bad things!")), BaleenError.__name__)

    def test_stype_embed_helper(self):
        """
        Test stype on reraises decorators.
        """

        @reraise(BaleenError)
        def badfunc():
            raise TypeError("This is clearly the wrong type!")

        try:
            badfunc()
        except BaleenError as e:
            self.assertEqual(stype(e), "BaleenError (TypeError)")

    def test_ingestor_hooks(self):
        """
        Test the started and finished ingestor hooks
        """

        # Create Ingestor and call the entry point method
        ingestor = get_ingest_mock()
        ingestor.ingest()

        # Assert that started and finished were called, and failed wasn't.
        ingestor.started.assert_called_once_with()
        ingestor.finished.assert_called_once_with()
        ingestor.failed.assert_not_called()

    def test_ingestor_failed_hook(self):
        """
        Test the started and failed ingestor hooks
        """

        ingestor = get_ingest_mock()
        ingestor.process.side_effect = Exception("Things went wrong!")

        # Call the entry point method
        with self.assertRaises(Exception) as cm:
            ingestor.ingest()

        # Assert that started and finished were called, and failed wasn't.
        ingestor.started.assert_called_once_with()
        ingestor.finished.assert_not_called()
        ingestor.failed.assert_called_once_with(cm.exception)

    def test_ingestor_state(self):
        """
        Ensure that the ingestor state is correctly modified
        """
        ingestor = get_ingest_mock()

        self.assertIsNone(ingestor.jobid)
        self.assertIsNone(ingestor.timer)

        ingestor.ingest()

        self.assertIsNotNone(ingestor.jobid)
        self.assertIsNotNone(ingestor.timer)



# baleen.console.commands.export
# Export utility to dump an HTML corpus to disk from the database.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 11:12:50 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: export.py [] benjamin@bengfort.com $

"""
Export utility to dump an HTML corpus to disk from the database.
"""

##########################################################################
## Imports
##########################################################################

import os
import baleen.models as db

from commis import Command
from baleen.config import settings
from baleen.console.utils import csv
from baleen.export import MongoExporter

##########################################################################
## Command
##########################################################################

class ExportCommand(Command):

    name = 'export'
    help = 'Export the raw HTML corpus for doing NLP'
    args = {
        ('-C', '--categories'): {
            'type': csv(str),
            'default': None,
            'help': 'Specify categories to export',
        },
        'location': {
            'nargs': 1,
            'type': str,
            'help': 'Location to write the corpus out to'
        }
    }

    def handle(self, args):
        # Connect to database
        db.connect()

        # Export from the database
        exporter = MongoExporter()
        exporter.export(args.location[0], categories=args.categories)

        # Perform counts of export
        root = args.location[0]
        cats = filter(
            os.path.isdir, [os.path.join(root, cat) for cat in os.listdir(root)]
        )
        docs = sum(len(os.listdir(d)) for d in cats)

        return (
            "Exported {} documents in {} categories "
            "as well as a readme to {}.".format(
                docs, len(cats), root
            )
        )



# baleen.feed
# Handles the synchronization of an RSS feed.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sun Sep 21 09:58:44 2014 -0400
#
# Copyright (C) 2014 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: feed.py [] benjamin@bengfort.com $

"""
Handles the synchronization of documents from an RSS feeds.
"""

##########################################################################
## Imports
##########################################################################

import feedparser

from baleen.models import Feed
from baleen.utils.timez import localnow
from baleen.exceptions import FeedTypeError
from baleen.exceptions import SynchronizationError
from baleen.utils.decorators import memoized, reraise


##########################################################################
## Module Constants
##########################################################################

FEEDPARSER_IGNORABLE_FIELDS = {
    'updated', 'updated_parsed', 'id',
    'published', 'published_parsed', 'category',
}


##########################################################################
## Feed Synchronization
##########################################################################

class FeedSync(object):
    """
    A utility that wraps both a Feed object and the feedparser library.
    The feed that is passed into the FeedSync can be one of the following:

        - a string representing the url to the RSS feed
        - a dictionary with an xmlUrl key (from OPML)
        - a Feed object loaded from MongoDB.

    The feed synchronization utility is smart enough to access what it needs.
    """

    URL   = "FEED_URL"
    DICT  = "FEED_DICT"
    MODEL = "FEED_MODEL"

    @classmethod
    def factory(klass, feeds):
        """
        Yields a feed synchronizer for each feed in the feeds.
        """
        for feed in feeds:
            yield klass(feed)

    def __init__(self, feed):
        """
        Feed can be a string (url), a dictionary with an `xmlUrl` or a Feed.
        """
        self.feed = feed

    @memoized
    def type(self):
        """
        Returns the type of the feed.
        """
        if isinstance(self.feed, basestring):
            return self.URL

        if isinstance(self.feed, Feed):
            return self.MODEL

        if isinstance(self.feed, dict):
            if not 'xmlUrl' in self.feed:
                raise FeedTypeError(
                    "Dictionary object does not contain 'xmlUrl' key!"
                )
            return self.DICT

        raise FeedTypeError(
            "Could not determine feed type from '{}'".format(type(self.feed))
        )

    @memoized
    def url(self):
        """
        Extracts the url from the feed based on the type.
        """
        return {
            self.URL: lambda: self.feed,
            self.DICT: lambda: self.feed.get('xmlUrl', None),
            self.MODEL: lambda: self.feed.link,
        }[self.type]()

    def parse(self):
        """
        Wraps the feedparser.parse function such that if the feed is an model,
        it uses the etag or modified to prevent duplicating the download.

        NOTE: Calling this function will NOT update the feed use sync instead!
        NOTE: Exceptions in this function will not be handled by Baleen!
        """
        # Only models contain the etag/modified saved information.
        if self.type == self.MODEL:
            # If there is an etag use it (even if there is also modified)
            if self.feed.etag:
                return feedparser.parse(self.url, etag=self.feed.etag)

            # If there is a modified date, then use it
            if self.feed.modified:
                return feedparser.parse(self.url, modified=self.feed.modified)

        # Otherwise just return the parse of the URL
        return feedparser.parse(self.url)

    @reraise(klass=SynchronizationError)
    def sync(self, save=True):
        """
        Calls the feedparser.parse function correctly but also synchronizes
        the state of the feed (e.g. last modified, etag, etc.) to MongoDB.

        Note: If the feed isn't a model, it just does the same as parse.

        If save is True (default) will save the Feed back to MongoDB.
        """
        # Get the result from the parse function.
        result = self.parse()

        # If this is not a model, bail out and return the result.
        if not self.type == self.MODEL:
            return result

        # Otherwise update the model in MongoDB with synchronization info.
        # Set the last fetched timestamp on the model.
        self.feed.fetched = localnow()

        # Update the feed properties from the result.
        for key in ('etag', 'modified', 'version'):
            if key in result and getattr(result, key):
                setattr(self.feed, key, getattr(result, key))

        # Update the link via the href
        if 'href' in result and result.href:
            self.feed.link = result.href

        # Update the feed items from the result.
        for key, val in result.feed.items():
            if key in FEEDPARSER_IGNORABLE_FIELDS:
                # Ignore these generated or protected fields.
                continue

            if key == 'link':
                self.feed.urls['htmlUrl'] = val

            elif key == 'links':
                for idx, link in enumerate(val):
                    if 'rel' in link:
                        self.feed.urls[link['rel'] + str(idx)] = link['href']
                    else:
                        self.feed.urls["link{}".format(idx)] = link['href']

            else:
                setattr(self.feed, key, val)

        if save:
            self.feed.save()

        return result

    def entries(self, save=True):
        """
        A helper function to simultaneously call sync and iterate over the
        entries from the feed. This is the usual method of interacting with
        the feed sync object. Note that this just returns raw dicts not Posts.
        """
        result = self.sync(save=save)
        return result.entries



# baleen.utils
# Utilities and helpers functions for the Baleen project.
#
# Author:   Benjamin Bengfort <bengfort@cs.umd.edu>
# Created:  Sun Feb 21 15:00:06 2016 -0500
#
# Copyright (C) 2016 University of Maryland
# For license information, see LICENSE.txt
#
# ID: __init__.py [] benjamin@bengfort.com $

"""
Utilities and helpers functions for the Baleen project.
"""

##########################################################################
## Imports
##########################################################################

# For the log configuration to work
from . import mongolog



# tests.test_feed
# Test the feed module - the main entry point to Baleen
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sun Feb 21 15:49:18 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: test_feed.py [] benjamin@bengfort.com $

"""
Test the feed module - the main entry point to Baleen
"""

##########################################################################
## Imports
##########################################################################

import os
import pickle
import unittest

from mongomock import MongoClient as MockMongoClient

try:
    from unittest import mock
except ImportError:
    import mock

from baleen.feed import *
from baleen.models import *
from urlparse import urlparse
from baleen.exceptions import FeedTypeError

##########################################################################
## Fixtures
##########################################################################

FIXTURES = os.path.join(os.path.dirname(__file__), "fixtures")
FEEDLY   = os.path.join(FIXTURES, "feedly.opml")
RESULT   = os.path.join(FIXTURES, "feedparser_result.pickle")

# Feed Fixtures
STR_FEED     = 'http://freakonomics.blogs.nytimes.com/feed/'
UNICODE_FEED = u'http://blog.kaggle.com/feed/'
OPML_FEED    = {
    "type":"rss", "text":"The Daily Notebook", "title":"The Daily Notebook",
    "xmlUrl":"https://mubi.com/notebook/posts.atom", "htmlUrl":"https://mubi.com/notebook/posts",
}
MONGO_FEED   = Feed(
    title = u'The Rumpus.net',
    link = u'http://therumpus.net/feed/',
    urls = {u'htmlurl': u'http://therumpus.net'},
    category = u'books',
)

##########################################################################
## Feed Synchronization Tests
##########################################################################

class FeedSyncTests(unittest.TestCase):

    def setUp(self):
        """
        Create the mongomock connection
        """
        self.conn = connect(host='mongomock://localhost')
        assert isinstance(self.conn, MockMongoClient)

        # Clear out the database
        for feed in Feed.objects(): feed.delete()
        for post in Post.objects(): post.delete()

    def tearDown(self):
        """
        Drop the mongomock connection
        """
        assert isinstance(self.conn, MockMongoClient)
        self.conn = None

    def test_fsync_factory(self):
        """
        Test multiple types in the feed sync factory
        """
        cases = (
            STR_FEED, UNICODE_FEED, OPML_FEED, MONGO_FEED
        )

        for fsync in FeedSync.factory(cases):
            self.assertIsInstance(fsync, FeedSync)

    def test_type_check(self):
        """
        Assert that strings, Feeds, and dicts can be sync'd
        """
        cases = (
            (STR_FEED, FeedSync.URL),
            (UNICODE_FEED, FeedSync.URL),
            (OPML_FEED, FeedSync.DICT),
            (MONGO_FEED, FeedSync.MODEL),
        )

        for feed, ftype in cases:
            fsync = FeedSync(feed)
            self.assertEqual(fsync.type, ftype)

    def test_bad_type(self):
        """
        Test that bad types raise an exception in sync
        """
        cases = (
            10, {u'htmlurl': u'https://mubi.com/notebook/posts'}, ['a','b','c']
        )

        for case in cases:
            fsync = FeedSync(case)
            with self.assertRaises(FeedTypeError):
                fsync.type

    def test_url_extraction(self):
        """
        Test the feed sync multiple type url extraction
        """
        cases = (
            (STR_FEED, STR_FEED),
            (UNICODE_FEED, UNICODE_FEED),
            (OPML_FEED, OPML_FEED['xmlUrl']),
            (MONGO_FEED, MONGO_FEED.link),
        )

        for feed, url in cases:
            fsync = FeedSync(feed)
            self.assertEqual(fsync.url, url)

    @mock.patch('baleen.feed.feedparser.parse')
    def test_feedparser_wrapping(self, mock_feedparser):
        """
        Test the feedparser access by mocking feedparser calls
        """

        # Ensure that the mocking worked out for us
        assert mock_feedparser is feedparser.parse

        cases = (
            (STR_FEED, STR_FEED),
            (UNICODE_FEED, UNICODE_FEED),
            (OPML_FEED, OPML_FEED['xmlUrl']),
            (MONGO_FEED, MONGO_FEED.link),
        )

        for feed, url in cases:
            fsync  = FeedSync(feed)
            result = fsync.parse()
            mock_feedparser.assert_called_with(url)

    @mock.patch('baleen.feed.feedparser.parse')
    def test_feedparser_wrapping(self, mock_feedparser):
        """
        Test etag and modified blocking on feedparser for Feed objects
        """

        # Ensure that the mocking worked out for us
        assert mock_feedparser is feedparser.parse

        feed = Feed(link = u'https://mubi.com/notebook/posts.atom')
        feed.etag = 'abcdefg'

        # Test Case 1: etag but no modified
        result = FeedSync(feed).parse()
        mock_feedparser.assert_called_with(feed.link, etag=feed.etag)

        # Test Case 2: modified but no etag
        feed.etag = None
        feed.modified = "Fri, 11 Jun 2012 23:00:34 GMT"
        result = FeedSync(feed).parse()
        mock_feedparser.assert_called_with(feed.link, modified=feed.modified)

        # Test Case 3: modified and etag
        feed.etag = 'hijklmnop'
        result = FeedSync(feed).parse()
        mock_feedparser.assert_called_with(feed.link, etag=feed.etag)

    @mock.patch('baleen.feed.feedparser.parse')
    def test_feed_sync(self, mock_feedparser):
        """
        Test that sync updates the Feed object
        """
        # Ensure that the mocking worked out for us
        assert mock_feedparser is feedparser.parse

        # Give the mock feedparser a result!
        with open(RESULT, 'rb') as f:
            mock_feedparser.return_value = pickle.load(f)

        fsync  = FeedSync(MONGO_FEED)
        result = fsync.sync()

        # Fetch the feed from the database.
        self.assertEqual(Feed.objects.count(), 1)
        feed = Feed.objects.first()

        # Ensure that the various properties have been set.
        self.assertEqual(feed.etag, u'W/"29e84abdc28e3fa87709d1f309b7c214-gzip"')
        self.assertEqual(feed.modified, u'Wed, 02 Mar 2016 22:00:06 GMT')
        self.assertEqual(feed.version, u'rss20')
        self.assertEqual(feed.link, MONGO_FEED.link)
        self.assertIsNotNone(feed.fetched)

    @mock.patch('baleen.feed.feedparser.parse')
    def test_feed_sync_mongodb(self, mock_feedparser):
        """
        Test the sync MongoDB interaction
        """
        # Ensure that the mocking worked out for us
        assert mock_feedparser is feedparser.parse

        # Give the mock feedparser a result!
        with open(RESULT, 'rb') as f:
            mock_feedparser.return_value = pickle.load(f)

        fsync  = FeedSync(MONGO_FEED)

        # Test sync without save
        result = fsync.sync(save=False)
        self.assertEqual(Feed.objects.count(), 0)

        # Test sync with save
        result = fsync.sync()
        self.assertEqual(Feed.objects.count(), 1)

    @mock.patch('baleen.feed.feedparser.parse')
    def test_feed_sync_non_model(self, mock_feedparser):
        """
        Test the sync with a non-model feed.
        """
        # Ensure that the mocking worked out for us
        assert mock_feedparser is feedparser.parse

        # Give the mock feedparser a result!
        with open(RESULT, 'rb') as f:
            mock_feedparser.return_value = pickle.load(f)

        fsync  = FeedSync(OPML_FEED)

        # Test sync without save
        result = fsync.sync()
        self.assertEqual(Feed.objects.count(), 0)



# tests.test_models
# Testing for the mongoengine models (basic stuff).
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Wed Mar 02 21:11:08 2016 -0500
#
# Copyright (C) 2016 Bengfort.com
# For license information, see LICENSE.txt
#
# ID: test_models.py [] benjamin@bengfort.com $

"""
Testing for the mongoengine models (basic stuff).
"""

##########################################################################
## Imports
##########################################################################

import unittest
import mongoengine as me

from mongomock import MongoClient as MockMongoClient

try:
    from unittest import mock
except ImportError:
    import mock

from baleen.models import *


##########################################################################
## Mongo Test Mixin
##########################################################################

class MongoTestMixin(object):

    def setUp(self):
        """
        Create the mongomock connection
        """
        self.conn = connect(host='mongomock://localhost')
        assert isinstance(self.conn, MockMongoClient)

        # Clear out the database
        for feed in Feed.objects(): feed.delete()
        for post in Post.objects(): post.delete()

    def tearDown(self):
        """
        Drop the mongomock connection
        """
        assert isinstance(self.conn, MockMongoClient)
        self.conn = None

    def assertDateTimeEqual(self, dta, dtb):
        """
        Assert that two datetimes are within 1 second of each other
        """
        dta = dta.replace(microsecond=0)
        dtb = dta.replace(microsecond=0)

        if dta.second != dtb.second:
            self.assertLessThanEqual(
                abs(dta.second - dtb.second), 1, "datetimes are not one second apart!"
            )
            dta = dta.replace(second=0)
            dtb = dtb.replace(second=0)

        self.assertEqual(dta, dtb)

##########################################################################
## Feed Model Tests
##########################################################################

class FeedModelTests(MongoTestMixin, unittest.TestCase):

    def test_link_requred(self):
        """
        Assert that the feed link is required
        """
        feed = Feed(title="My Awesome Feed", category="socks")
        with self.assertRaises(me.ValidationError):
            feed.save()

    def test_created_updated(self):
        """
        Ensure the feed updated timestamp is tracked
        """
        feed = Feed(title="A News Feed", category="news", link="https://example.com/feed.atom")
        feed.save()

        self.assertIsNotNone(feed.created)
        self.assertIsNotNone(feed.updated)
        self.assertDateTimeEqual(feed.created, feed.updated)

        feed.title = "An Olds Feed"
        feed.save()
        self.assertNotEqual(feed.created, feed.updated)

    def test_properties(self):
        """
        Test the properties of the feed model
        """
        feed = Feed(title="A News Feed", category="news", link="https://example.com/feed.atom")
        feed.save()

        self.assertEqual(feed.xmlurl, feed.link)
        self.assertIsNone(feed.htmlurl)

        feed.urls = {'htmlUrl': 'https://example.com/'}
        feed.save()

        self.assertEqual(feed.htmlurl, 'https://example.com/')

    def test_stringify(self):
        """
        Test the stringification of a feed
        """
        feed = Feed(category="news", link="https://example.com/feed.atom")
        feed.save()

        self.assertEqual(str(feed), feed.link)

        feed.title = "A News Feed"
        feed.save()

        self.assertEqual(str(feed), feed.title)


##########################################################################
## Post Model Tests
##########################################################################

class PostModelTests(MongoTestMixin, unittest.TestCase):

    def test_url_requred(self):
        """
        Assert that the post url is required
        """
        post = Post(title="My Awesome Post", content="socks")
        with self.assertRaises(me.ValidationError):
            post.save()

    def test_created_updated(self):
        """
        Ensure the post updated timestamp is tracked
        """
        post = Post(title="My Awesome Post", content="socks", url="http://example.com/socks.html")
        post.save()

        self.assertIsNotNone(post.created)
        self.assertIsNotNone(post.updated)
        self.assertDateTimeEqual(post.created, post.updated)

        post.title = "My even more awesome Post!"
        post.save()
        self.assertNotEqual(post.created, post.updated)

    def test_content_hashing(self):
        """
        Test the automatic hashing of content
        """
        post = Post(content="socks", url="http://example.com/socks.html")
        self.assertIsNone(post.signature)
        post.save()

        self.assertIsNotNone(post.signature)
        self.assertEqual(post.signature, '54f6d9fbe8ee576f82d6eb7e4d1d55691a1f0b7bd956246d3de56ee84bd1d333')

    def test_stringify(self):
        """
        Test the stringification of a post
        """
        post = Post(content="socks", signature="abc", url="http://example.com/socks.html")
        post.save()

        self.assertEqual(str(post), post.url)

        post.title = "My Awesome Post"
        post.save()

        self.assertEqual(str(post), post.title)
